from neuroiatools.DisplayData.plotEEG import plotEEG
from neuroiatools.SignalProcessor import Filter
import h5py
import pandas as pd

eeg_data = h5py.File("datasets\\raweeg_executed_tasks.hdf5", "r")["raw_eeg"][:]
print("raweeg shape: ", eeg_data.shape)
eventos = pd.read_csv("datasets\\events_executed_tasks.txt")
print(eventos.head())

sfreq = 512

event_times = eventos["event_time"].values/sfreq ##marcado de eventos en segundos
event_labels = eventos["class_name"].values

##grafico el EEG sin filtrar desde el segundo 45
plotEEG(eeg_data, sfreq, event_times, event_labels, scalings = 40,show=True, block=True,
        duration = 30, start = 45, remove_dc = True, bad_color = "red")

##filtramos el EEG y luego graficamos
filtro = Filter.Filter(lowcut=0.5, highcut=38, notch_freq=50.0, notch_width=2.0, sample_rate=512.0)
filtered_eeg = filtro.filter_data(eeg_data)

##grafico el EEG sin filtrar desde el segundo 45
plotEEG(filtered_eeg, sfreq, event_times, event_labels, scalings = 40,show=True, block=True,
        duration = 30, start = 45, remove_dc = True, bad_color = "red")


from neuroiatools.SignalProcessor import Filter
from neuroiatools.EEGManager import TrialsHandler
import h5py
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

##cargo datasets/raweeg_executed_tasks.hdf5
raweeg = h5py.File("datasets\\raweeg_executed_tasks.hdf5", "r")["raw_eeg"][:]
print("raweeg shape: ", raweeg.shape)
eventos = pd.read_csv("datasets\\events_executed_tasks.txt")
print(eventos.head())

sfreq = 512

trial = 2
tarea = eventos.loc[1,"class_name"]
canal = 32

time_events = eventos["event_time"].values/sfreq
trial_time = time_events[trial-1]

tmin = -2
tmax = 4

t_fi = 0.5 #tiempo fade-in
t_cue = 2 #tiempo cue
t_fo = 1 #tiempo fade-out

total_time = t_fi + t_cue + t_fo

##instancio un objeto de la clase TrialsHandler
th = TrialsHandler.TrialsHandler(raweeg, time_events, sfreq, tmin=tmin, tmax=tmax, reject=None)
trials = th.trials #obtengo los trials

ejet = np.arange(0, trials.shape[2])/sfreq

##creo un objeto Filter
filtro = Filter.Filter(lowcut=7, highcut=28, notch_freq=50.0, notch_width=2.0, sample_rate=512.0)

##aplico el filtro a los trials
trials_filtered = filtro.filter_data(trials)

fig, ax = plt.subplots(figsize=(15, 4))
plt.plot(ejet+trial_time+tmin, trials_filtered[trial-1,canal-1,:], color = "#28b463",label="Señal Filtrada")
plt.plot(ejet+trial_time+tmin, trials[trial-1,canal-1,:], color = "#34495e", label="Señal Original", alpha=0.6, linestyle="--")
##agrego sombras
plt.axvspan(trial_time-t_fi, trial_time, alpha=0.2, color='gray',label='Fade-in')
plt.axvspan(trial_time, trial_time+t_cue, alpha=0.2, color='yellow',label='Cue')
plt.axvspan(trial_time+t_cue, trial_time+t_cue+t_fo, alpha=0.2, color='#3498db',label='Fade-out')
##linea de cue
plt.axvline(trial_time, color='black', linestyle=':', label='Cue on')
plt.xlabel("Tiempo [s]")
plt.ylabel("Amplitud [uV]")
plt.legend(loc="upper right")
plt.title("EEG original y filtrado" + " - Trial "+str(trial)+" - "+tarea + " - Canal "+str(canal))
plt.grid()
plt.show()


import numpy as np
import mne

def plotEEG(eeg_data, sfreq, event_times, event_labels, channel_names=None,
            title = "EEG", scalings=dict(eeg=40e-6), color = {"eeg":"blue"}, bgcolor = "#eaeded", **kwargs):
    """
    Genera una gráfica interactiva para analizar un registro de EEG con eventos marcados.

    Utiliza la librería MNE para generar la gráfica interactiva.

    Parámetros:
    - eeg_data (np.ndarray): Array de forma [canales, muestras] con los datos de EEG.
    - sfreq (float): Frecuencia de muestreo en Hz.
    - event_times (list or np.ndarray): Momentos en segundos donde ocurrieron los eventos.
    - event_labels (list): Nombres de los eventos, deben coincidir en longitud con `event_times`.
    - channel_names (list, opcional): Nombres de los canales. Si no se proporcionan, se generarán automáticamente.
    - title (str, opcional): Título de la gráfica.
    - scalings (dict): Escalas para los canales. Por defecto, 'eeg'=40e-6. Puede ser "auto" o bien "None".
    - color (dict): Especifíca el color de los trazos del EEG. Por defecto es azul.
    - bgcolor: Color del fondo.
    - kwargs: Argumentos adicionales para la función `mne.io.Raw.plot`. Ver en https://mne.tools/stable/generated/mne.io.Raw.html#mne.io.Raw.plot

    Retorna:
    - None. Muestra una gráfica interactiva.
    """
    if eeg_data.ndim != 2:
        raise ValueError("eeg_data debe ser un array de forma [canales, muestras].")

    if len(event_times) != len(event_labels):
        raise ValueError("La cantidad de event_times debe coincidir con la cantidad de event_labels.")

    n_channels, n_samples = eeg_data.shape
    duration = n_samples / sfreq  # Duración del registro en segundos.

    ## Generamos nombres de canales si no se proporcionaron
    if channel_names is None:
        channel_names = [f"Ch{i+1}" for i in range(n_channels)]

    ## Creamos un objeto mne.Info con la información de los canales
    info = mne.create_info(ch_names=channel_names, sfreq=sfreq, ch_types='eeg')

    ## Creamos un objeto mne.RawArray con los datos de EEG
    raw = mne.io.RawArray(eeg_data, info)

    ## Creamos un objeto mne.Annotations con los eventos
    annotations = mne.Annotations(onset=event_times,  ## Eventos en segundos
                                   duration=[0] * len(event_times), 
                                   description=event_labels)
    
    raw.set_annotations(annotations)
    
    ##Graficamos
    raw.plot(scalings=scalings,
             title=title,
             color = color, bgcolor = bgcolor,
             **kwargs)

if __name__ == "__main__":

    # Ejemplo con datos ficticios
    n_channels = 64
    sfreq = 512  # Frecuencia de muestreo en Hz
    duration_sec = 900  # Duración de 5 minutos
    n_samples = int(sfreq * duration_sec)

    # Generar datos ficticios (ruido aleatorio)
    np.random.seed(42)
    eeg_data = np.random.randn(n_channels, n_samples)*20e-6

    # Eventos ficticios
    event_times = [10, 50, 120, 200, 800]  # En segundos
    event_labels = ['Inicio', 'Tarea 1', 'Tarea 2', 'Pausa', 'Fin']

    # Llamar a la función
    plotEEG(eeg_data, sfreq, event_times, event_labels, scalings = 100e-6,show=True, block=True,
            duration = 20, start = 0)


import numpy as np
from scipy.signal import butter, filtfilt, iirnotch
from sklearn.base import BaseEstimator, TransformerMixin

class Filter(BaseEstimator, TransformerMixin):
    """Clase para filtrar señales provenientes de la placa openBCI. Las señales llegan en un numpy array.
    La idea es aplicar un filtro pasa banda y un filtro notch a la señal a todo el array.
    El filtrado se aplica sobre el último axis."""

    def __init__(self, lowcut = 8.0, highcut = 38.0, notch_freq = 50.0, notch_width = 2.0, sample_rate = 512.0,
                 padlen = None, order = 4, discard_samples = 0):
        """Inicializa el objeto con los parámetros de filtrado.
        -lowcut: Frecuencia de corte inferior del filtro pasa banda.
        -highcut: Frecuencia de corte superior del filtro pasa banda.
        -notch_freq: Frecuencia de corte del filtro notch.
        -notch_width: Ancho de banda del filtro notch.
        -sample_rate: Frecuencia de muestreo de la señal.
        -X: Señal de entrada. No se usa en este caso."""

        self.lowcut = lowcut
        self.highcut = highcut
        self.notch_freq = notch_freq
        self.notch_width = notch_width
        self.sample_rate = sample_rate
        self.padlen = padlen
        self.order = order
        self.discard_samples = int(discard_samples*self.sample_rate) #muestras a descartar al inicio de la señal filtrada

        self.b, self.a = butter(self.order, [self.lowcut, self.highcut], btype='bandpass', fs=self.sample_rate)
        self.b_notch, self.a_notch = iirnotch(self.notch_freq, 20, self.sample_rate)

    def filter_data(self, signal):
        """Función para aplicar los filtros a la señal.
        -signal: Es la señal en un numpy array de la forma [n-trials, canales, muestras]."""

        signal = signal - np.mean(signal, axis=-1, keepdims=True)
        signal = filtfilt(self.b, self.a, signal, axis = -1, padlen = self.padlen) #aplicamos el filtro pasa banda
        signal = filtfilt(self.b_notch, self.a_notch, signal, axis = -1, padlen = self.padlen) #aplicamos el filtro notch
        return signal[..., self.discard_samples:] #descartamos las primeras muestras

if __name__ == "__main__":

    pass


import requests
from pathlib import Path

def download_data(file_name, save_dir="datasets", repo_url="https://raw.githubusercontent.com/lucasbaldezzari/neuroiatools/main/datasets"):
    """
    Descarga un archivo de datos desde un repositorio de GitHub y lo guarda localmente,
    mostrando el progreso de descarga.
    
    Parameters:
        file_name (str): Nombre del archivo que deseas descargar (e.g., "data1.npy").
        save_dir (str): Directorio local donde se guardará el archivo.
        repo_url (str): URL base del repositorio donde están los datos.
        
    Returns:
        str: Ruta completa del archivo descargado.
    """
    # Crear el directorio si no existe
    save_path = Path(save_dir)
    save_path.mkdir(parents=True, exist_ok=True)
    
    # URL completa del archivo
    file_url = f"{repo_url}/{file_name}"
    local_file_path = save_path / file_name
    
    # Descargar solo si no existe localmente
    if not local_file_path.exists():
        print(f"Descargando {file_name} desde {file_url}...")
        response = requests.get(file_url, stream=True)
        if response.status_code == 200:
            # Obtener tamaño total del archivo si está disponible
            total_size = int(response.headers.get('Content-Length', 0))
            total_size_mb = total_size / (1024 * 1024) if total_size else 0
            
            with open(local_file_path, "wb") as f:
                downloaded_size = 0
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
                    downloaded_size += len(chunk)
                    
                    # Calcular progreso
                    if total_size > 0:
                        percent = (downloaded_size / total_size) * 100
                        downloaded_mb = downloaded_size / (1024 * 1024)
                        print(f"\rDescargando: {downloaded_mb:.2f} MB / {total_size_mb:.2f} MB ({percent:.2f}%)", end="")
            print(f"\nArchivo guardado en {local_file_path}")
        else:
            raise FileNotFoundError(f"No se pudo descargar el archivo: {file_url}. Status code: {response.status_code}")
    else:
        print(f"El archivo {file_name} ya existe en {local_file_path}.")
    
    return str(local_file_path)


from setuptools import setup, find_packages
import toml
import os

def getRequirements():
    with open('requirements.txt', 'r') as f:
        requirements = f.read().splitlines()
    return requirements

def readme():
    with open('README.md') as f:
        return f.read()
    
def version():
    with open('pyproject.toml') as f:
        return toml.load(f)['project']['version']

# Encuentra automáticamente los archivos en 'neuroiatools/datasets'
def find_package_data():
    package_data = {}
    base_path = os.path.join("neuroiatools", "datasets" )
    for root, _, files in os.walk(base_path):
        rel_root = os.path.relpath(root, "neuroiatools")
        package_data[rel_root.replace(os.path.sep, ".")] = files
    return package_data

setup(
    name='neuroiatools',
    version=version(),
    packages=find_packages(),
    install_requires=getRequirements(),
    include_package_data=True,  # Habilita la inclusión de datos no Python
    package_data=find_package_data(),  # Agrega los archivos dentro de datasets
    author='LUCAS BALDEZZARI',
    author_email='lmbaldezzari@gmail.com',
    description='Tools for EEG processing and analysis',
    long_description=readme(),
    long_description_content_type='text/markdown',  # Para renderizar markdown en PyPI
    url='https://github.com/lucasbaldezzari/neuroiatools',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 11',
        'Operating System :: Unix',
    ],
)


import numpy as np
import logging
import pandas as pd

class TrialsHandler():
    """Clase para obtener los trials a partir de raw data"""

    def __init__(self, rawEEG, timeEvents, sfreq, tmin = -0.2, tmax = 1, reject = None, trialsToRemove = None) -> None:
        """Constructor de la clase Trials
        Parametros:
            - rawEEG (numpy.array): array de numpy con la señal de EEG de la forma [channels, samples]
            - timeEvents (numpy.array): array de numpy con los eventos de tiempo en segundos.
            - sfreq (float): frecuencia de muestreo de la señal de EEG.
            - tmin, tmax: tiempo inicial del trial y tiempo final del trial relativos al timeEvent en segundos.
            - reject (float): Valor de umbral para rechazar trials. Si el valor absoluto Pico a Pico o (Peak to Peak PTP) de alguno de los canales
            supera este valor, el trial es rechazado. Si es None, no se rechazan trials.
            - trialsToRemove (list): lista de trials a remover. Si es None, no se remueven trials.
            """
        
        if not isinstance(rawEEG, np.ndarray):
            raise TypeError("rawEEG debe ser un array de numpy")
        self.rawEEG = rawEEG
        if not isinstance(timeEvents, np.ndarray):
            raise TypeError("timeEvents debe ser un array de numpy")
        self.timeEvents = timeEvents
        self.sfreq = sfreq
        self.tmin = tmin
        self.tmax = tmax
        
        self.trials = self.getTrials() #array de numpy con los trials de la forma [trials, channels, samples]
        #chequeamos si hay trials que remover
        if trialsToRemove is not None:
            self.removeTrials(trialsToRemove)

        self.rejectedTrials = None
        self.channels_exceeded_ptp = None
        self.reject = reject
        if self.reject is not None:
            self.rejectedTrials, self.channels_exceeded_ptp = self._rejectTrials()

    def getTrials(self):
        """Función para extraer los trials dentro de self.rawEEG"""

        if self.tmin > self.tmax:
            raise ValueError("tmin debe ser menor que tmax")

        init_idx = np.round((self.timeEvents + self.tmin) * self.sfreq).astype(int) #índices de inicio de cada trial
        end_idx = np.round((self.timeEvents + self.tmax) * self.sfreq).astype(int) #índices de fin de cada trial
        ## concatenamos en time_idx los índices de inicio y fin de cada trial
        time_idx = np.vstack((init_idx, end_idx)).T  # transponemos para que tenga forma (n_trials, 2)

        # Tamaño del segmento esperado
        segment_length = int(round((self.tmax - self.tmin) * self.sfreq)) ##número de muestras en el segmento

        trials = np.zeros((len(init_idx), self.rawEEG.shape[0], segment_length)) ##array de numpy con los trials de la forma [trials, channels, samples]

        for i, (start, end) in enumerate(time_idx):
            if end - start != segment_length:
                raise ValueError(
                    f"El tamaño del segmento ({end - start}) no coincide con el tamaño esperado ({segment_length}) "
                    f"para el trial {i}."
                )
            trials[i] = self.rawEEG[:, start:end]

        return trials
    
    def saveTrials(self, filename):
        """Función para guardar los trials en un archivo .npy"""
        np.save(filename, self.trials)
        print("Se han guardado los trials en {}".format(filename))

    def removeTrials(self, trialsToRemove:list):
        """Función para remover trials usando la lista de trialsToRemove.
        A partir de trialsToRemove removemos los indices de self.timeEvents, luego actualiamos
        self.trials y self.labels"""

        ##chequeamos que trialsToRemove sea una lista
        if not isinstance(trialsToRemove, list):
            raise TypeError("trialsToRemove debe ser una lista")
        
        ##chequeamos que los valores de los trials existan cómo indices
        if not all(trial in self.timeEvents.index for trial in trialsToRemove):
            raise ValueError("Uno o más valores pasados como trials no existen cómo índices en self.timeEvents")
        
        else:
            #removemos los trials de self.timeEvents
            self.timeEvents = self.timeEvents.drop(trialsToRemove)
            #eliminamos los trials de self.trials
            self.trials = np.delete(self.trials, trialsToRemove, axis=0)
            #eliminamos los trials de self.labels
            self.labels = np.delete(self.labels, trialsToRemove, axis=0)

            print("Se han removido los trials {}".format(trialsToRemove))

    def _rejectTrials(self):
        """
        Rechaza los trials en base a la amplitud pico a pico (PTP) máxima permitida.

        Para cada trial, calcula la amplitud PTP para cada canal. Si el PTP de algún canal en un trial
        excede el valor de umbral definido en `self.reject`, ese trial es rechazado.

        La idea es sacada de https://mne.tools/stable/generated/mne.Epochs.html

        Retorna:
            list: Lista de índices de los trials rechazados.
        """

        rejected_trials = []  # Lista para almacenar los índices de trials rechazados
        channels_exceeded_ptp = [] ## Lista para almacenar los canales que excedieron el umbral de PTP

        #recorremos los trials
        for trial_idx, trial in enumerate(self.trials):
            ptp_values = trial.max(axis=1) - trial.min(axis=1)
            exceeded_channels = np.where(ptp_values > self.reject)[0]
            if len(exceeded_channels) > 0:
                rejected_trials.append(trial_idx)
                channels_exceeded_ptp.append(exceeded_channels.tolist())

        # Remover los trials rechazados de self.trials
        self.trials = np.delete(self.trials, rejected_trials, axis=0)

        ## Actualizamos la lista de eventos de tiempo (self.timeEvents) eliminando los rechazados
        self.timeEvents = np.delete(self.timeEvents, rejected_trials, axis=0)

        print(f"Se han rechazado los siguientes trials por exceder el umbral de PTP: {rejected_trials}")

        return rejected_trials, channels_exceeded_ptp
    
if __name__ == "__main__":
    import numpy as np
    import matplotlib.pyplot as plt
    ##eeg de prueba
    sfreq = 512
    rawEEG = rawEEG = np.load("datasets\\signal_test_trialshandler.npy")
    timevents = np.load("datasets\\timevents_signal_test_trialshandler.npy")

    ejet = np.arange(0,len(rawEEG[0]))/sfreq
    plt.plot(ejet, rawEEG[0], label="Señal de prueba")
    ##agrego linea vertical en el tiempo del evento
    plt.axvline(x=timevents[0], color='r', linestyle='--', label="Evento")
    plt.title("Señal de prueba")
    plt.legend()
    plt.show()
    
    tmin = -1
    tmax = 3
    trialshandler = TrialsHandler(rawEEG, timevents, sfreq, tmin=tmin, tmax=tmax, reject=3)

    trials = trialshandler.trials
    ejet = np.arange(0,trials.shape[2])/sfreq+timevents[0]+tmin
    plt.plot(ejet,trials[0,0,:], label="Trial extraído")
    ##agrego una línea vertical en el tiempo del evento
    plt.axvline(x=timevents[0], color='r', linestyle='--', label="Evento")
    ##agrego una línea vertical en tmin
    plt.axvline(x=timevents[0]+tmin, color='g', linestyle='--', label="tmin")
    ##agrego una línea vertical en tmax
    plt.axvline(x=timevents[0]+tmax, color='y', linestyle='--', label="tmax")
    plt.title("Trial extraído")
    plt.legend()
    plt.show()
    


from neuroiatools.utils.datasetManager import download_data
import os
##obtengo la carpeta actual
current_dir = os.getcwd()
download_data(file_name = "events_executed_tasks.txt", save_dir=current_dir+"\\datasets")